/*  upacct.c : ver 1.6 : 06/04/95  */

/* 
 *  Copyright 1995, University of Liverpool, United Kingdom
 *
 *  Author: d.j.nixon@csc.liv.ac.uk
 */

/*
 *  Usage: upacct <input/output 'upacct' file> [input 'upacct' file(s) ...]
 *                                          [-p input 'pacct' files(s) ...]
 *
 *   Creates 'upacct' format files of per-user command usage records. 
 *  Input data is taken from process accounting 'pacct' files and/or existing
 *  'upacct' files. Data from special users - such as 'root' or 'adm' - is 
 *  ignored. When 'pacct' files are input the hostname is recorded with each
 *  record.
 *  
 *   Records are stored in most recent first order: There is no particular
 *  order between groups of user records.
 *
 ************************************************************************** 
 **************************************************************************
 *  'upacct' file format: header,record,record...,header,record ...       *
 *									  *
 *	header:		record count 				          *
 *			UID						  *
 *                       (field)                    			  *	
 *	record:		begin time	    |                             *
 *			elapsed time	    |> Per-user information       *
 *			command name	    |                             *
 *                      host name                                         *
 **************************************************************************
 *
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/acct.h>
#include <sys/param.h>

#include <netinet/in.h>
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>

/* global data */
extern char *sys_errlist[];

/* list of per user data */
typedef struct uacct {
  time_t ac_btime;       /* Beginning time */
  comp_t ac_etime;       /* acctng elapsed time in clock ticks */
  char ac_comm[8];       /* command name */
  struct in_addr addr;   /* host name */
  struct uacct *next;
} uacct;

/* upacct file record header */
typedef struct uahdr {
u_short uid;
unsigned count;
} uahdr;

/* list of users */
typedef struct ualist {
  u_short uid;
  uacct *data;
  uacct *data_end;
  struct ualist *next;
} ualist;

main(argc, argv)
int argc;
char **argv;
{

#define USER_UIDS 250    /* data is gathered from UIDS above this value */
#define MAXUSERS 2000
#define MAXURECS 1000000
#define MALLOC(els, mem) els-- ? mem++ : 0 

struct acct acctbuf;
uahdr hdr;
register int bytes, pacct, upacct, n=argc;
register ualist *user_list=NULL,*user_list_end,*user,*ul;
register uacct  *cmd,*ud,*uacctbuf=(uacct*)&acctbuf;
int pflag=0;   /* flag set if pacct files are specified */
register unsigned ul_elements=MAXUSERS,
                  ud_elements=MAXURECS;
ualist *ul_element;
uacct  *ud_element;

  /* parse command line */
  if (argc < 2)
  {
    fputs("Usage: upacct <output upacct file> [input upacct file(s) ...] [-p input pacct files(s) ...]\n",stderr);
    exit(1);
  }
  while (--n)
    if (!strcmp("-p", argv[n])) 
      if ((n < argc - 1) && (n > 1))
      {
        ++pflag;
        break;
      }
      else
      {
        fputs("Usage: upacct <output upacct file> [input upacct file(s) ...] [-p input pacct files(s) ...]\n",stderr);
        exit(1);
      } 

  /* pre-allocate memory */
  while(ul_elements)
  {
    if (ul_element=malloc(ul_elements*sizeof(ualist))) break;
    ul_elements-=ul_elements/2;
  }
  while(ud_elements)
  {
    if (ud_element=malloc(ud_elements*sizeof(uacct))) break;
    ud_elements-=ud_elements/2;
  }

  /* create output file */
  if ((upacct=open(argv[1], O_CREAT | O_RDWR, 0600)) == -1)
  {
    fprintf(stderr,"upacct: can't open %s: %s\n",argv[1],sys_errlist[errno]);
    exit(1);
  }

  /*  Input data from 'upacct' file(s): first file to last.
   */
  {
    for(n=1; n < argc; ++n)
    {
      if (!strcmp("-p", argv[n])) break;
      if ((pacct=open(argv[n], O_RDONLY)) == -1)
      {
        fprintf(stderr,"upacct: can't open %s: %s\n",argv[n],sys_errlist[errno]);
        exit(1);
      }
      while(bytes=read(pacct, &hdr, sizeof(uahdr)))
      {
        if (bytes == -1)
        {
          fprintf(stderr,"upacct: error reading %s: %s\n",argv[n],sys_errlist[errno]);
          exit(1);
        }
        user=NULL; ul=user_list;
        if (n != 1)              /* first file: don't search list */
          while(ul)              /* search list of users */
          {
            if (ul->uid == hdr.uid)
            {
              user=ul;
              break;
            }
            user_list_end=ul;
            ul=ul->next;
          }
        if (!user)               /* create a user entry */
        {
          if ((user=MALLOC(ul_elements, ul_element)) == NULL)
            if ((user=(ualist*)malloc(sizeof(ualist))) == NULL)
            {
              fputs("upacct: definitely out of memory\n",stderr);
              exit(2);
            }
          user->uid=hdr.uid;
          user->data=NULL;
          user->next=NULL;
          if(!user_list)
          {
            user_list=user_list_end=user;   /* start list of users */
          }
          else
          {
            user_list_end->next=user;       /* link in user */
            user_list_end=user;
          }
        }
        while(hdr.count--)
        {
          if ((cmd=MALLOC(ud_elements, ud_element)) == NULL)
            if ((cmd=(uacct*)malloc(sizeof(upacct))) == NULL)
            {
              fputs("upacct: out of memory\n",stderr);
              exit(2);
            }
          if (read(pacct, uacctbuf, (sizeof(uacct) - sizeof(uacct*))) == -1)
          {
            fprintf(stderr,"upacct: error reading %s: %s\n",argv[n],sys_errlist[errno]);
            exit(1);
  
          }
          cmd->ac_btime=uacctbuf->ac_btime;
          cmd->ac_etime=uacctbuf->ac_etime;
          memcpy(&cmd->ac_comm, &uacctbuf->ac_comm, 8);
          cmd->addr=uacctbuf->addr;
          cmd->next=NULL;
          if (user->data)
          {
            if (cmd->ac_btime >= user->data->ac_btime)
            {
              cmd->next=user->data;
              user->data=cmd;             /* prepend to list */
            }
            else if (cmd->ac_btime <=  user->data_end->ac_btime)
            {
              user->data_end->next=cmd;   /* append to list */
              user->data_end=cmd;
            }
            else                          /* insert in list */
            {
              ud=user->data;
              do
              {
                if (cmd->ac_btime >= ud->next->ac_btime)
                {
                  cmd->next=ud->next;
                  ud->next=cmd;
                  break;
                }
              }
              while (ud=ud->next);
            }
          }
          else   /* empty list */
          {
            user->data=cmd;
            user->data_end=cmd;
          }
        }
      }
      close(pacct);
    }
  }

  /*  Input selected data from 'pacct' file(s): last file to first.
   */
  if (pflag)
  {
  struct in_addr myaddr;
    /* get the local IP address for inclusion with the data */
    {
    char hname[MAXHOSTNAMELEN+1];
    struct hostent  *hp;
      /* get the local host name */
      if (gethostname(hname, MAXHOSTNAMELEN) == -1)
      {
        perror("upacct: don't know my host name!");
        exit(1);
      }
      if ((hp=gethostbyname(hname)) == NULL)
      {
        perror("upacct: can't get my host address!");
        exit(1);
      }
      /* save it */
      myaddr.s_addr=((struct in_addr *)(hp->h_addr))->s_addr;
    }
    while (--argc > 2)
    {
      if (!strcmp("-p", argv[argc])) break;
      if ((pacct=open(argv[argc], O_RDONLY)) == -1)
      {
        fprintf(stderr,"upacct: can't open %s: %s\n",argv[argc],sys_errlist[errno]);
        exit(1);    
      }
      while(bytes=read(pacct, &acctbuf, sizeof(struct acct)))
      {
        if (bytes != sizeof(struct acct))
          if (bytes == -1)
          {
            fprintf(stderr,"upacct: error reading %s: %s\n",argv[argc],sys_errlist[errno]);
            exit(1);
          }
          else
          {
            fprintf(stderr,"upacct: short read of %s\n",argv[argc]);
            exit(1);
          }
        user=NULL;ul=user_list;
        if (acctbuf.ac_uid < USER_UIDS)   /* don't bother with system user data */
          continue;
        while (ul)             /* search list for user */
        { 
          if (ul->uid == acctbuf.ac_uid) 
          {
            user=ul;
            break;
          }
          if (!ul->next) break;
          ul=ul->next;
        } 
        if (!user)            /* user not in list */
        {
          if ((user=MALLOC(ul_elements, ul_element)) == NULL)
            if ((user=(ualist*)malloc(sizeof(ualist))) == NULL)
            {
              fputs("upacct: out of memory\n",stderr);
              exit(2);
            }
          user->uid=acctbuf.ac_uid;
          user->data=NULL;
          user->next=NULL;
          if(!user_list)
            user_list=user;   /* start list of users */
          else
            ul->next=user;    /* link in user */
        }
        /* store command data */
        if ((cmd=MALLOC(ud_elements, ud_element)) == NULL)
          if ((cmd=(uacct*)malloc(sizeof(upacct))) == NULL)
          {
            fputs("upacct: out of memory\n",stderr);
            exit(2);
          }
        cmd->ac_btime=acctbuf.ac_btime;
        cmd->ac_etime=acctbuf.ac_etime;
        memcpy(&cmd->ac_comm, &acctbuf.ac_comm, 8);
        cmd->addr=myaddr;
        cmd->next=NULL;
        if (user->data)
        {
          if (cmd->ac_btime >= user->data->ac_btime)
          {
            cmd->next=user->data;
            user->data=cmd;             /* prepend to list */
          }
          else  if (cmd->ac_btime <=  user->data_end->ac_btime)
          {
            user->data_end->next=cmd;   /* append to list */
            user->data_end=cmd;
          }
          else
          {
            ud=user->data;              /* insert in list */
            do
            {
              if (cmd->ac_btime >= ud->next->ac_btime)
              {
                cmd->next=ud->next;
                ud->next=cmd;
                break;
              }
            }
            while (ud=ud->next);
          }
        }
        else   /* empty list */
        {
          user->data_end=cmd;
          user->data=cmd;
        }
      }
      close(pacct);
    }
  }

  /*  Write memory data to output file. 
   */
  {
  register unsigned rec_cnt=0;
    if (!user_list)
    {
      fputs("upacct: no data found\n",stderr);
      exit(1); 
    }
    do
    {
      hdr.uid=user_list->uid;
      bytes=0;
      ud=user_list->data;
      do {++bytes;} while (ud=ud->next);
      hdr.count=bytes;
      if (write(upacct, &hdr, sizeof(uahdr)) == -1)
      {
        fprintf(stderr,"upacct: error writing %s: %s\n",argv[1],sys_errlist[errno]);
        exit(1);
      }
      ud=user_list->data;
      do 
      { 
        if (write(upacct, ud, (sizeof(uacct) - sizeof(uacct*))) == -1)
        {
          fprintf(stderr,"upacct: error writing %s: %s\n",argv[1],sys_errlist[errno]);
          exit(1);
        }
        ++rec_cnt;
      } while (ud=ud->next);
    } while(user_list=user_list->next);
    printf("* user commands: %u\n",rec_cnt);
  }
  
  /* finis */
  exit(0);
}
